(***********************************************************************)
(*                                                                     *)
(*                           FaCiLe                                    *)
(*                 A Functional Constraint Library                     *)
(*                                                                     *)
(*            Nicolas Barnier, Pascal Brisset, LOG, CENA               *)
(*                                                                     *)
(* Copyright 2001 CENA. All rights reserved. This file is distributed  *)
(* under the terms of the GNU Lesser General Public License.           *)
(***********************************************************************)
(* $Id: fcl_reify.ml,v 1.14 2001/06/19 11:00:55 barnier Exp $ *)

open Fcl_var

class reification c b on_not = object
  inherit Fcl_cstr.t Fcl_cstr.normal
  method name = "reification"
  method fprint s = Printf.fprintf s "reification: "; c#fprint s
  method delay x =
    c#delay x;
    if on_not then c#not#delay x;
    delay [Attr.on_subst] b x
  method update =
    match Fd.value b with
      Val vb ->
	Fcl_stak.set solved true;
	Fcl_cstr.post (if vb = 0 then c#not else c)
    | Unk _ ->
	try
	  if c#is_solved || c#check then begin
	    Fcl_stak.set solved true;
	    Fd.unify b 1
	  end else begin
	    Fcl_stak.set solved true;
	    Fd.unify b 0
	  end
	with
	  Fcl_cstr.DontKnow -> ()
end

let cstr ?(delay_on_negation = true) c b =
  new reification c b delay_on_negation;;

let boolean ?delay_on_negation c  =
  let b = Fd.create Fcl_domain.boolean in
  let r = cstr ?delay_on_negation c b in
  Fcl_cstr.post r;
  b;;

exception MyDontKnow;;

let rec (||~~) c1 c2 =
  let update () =
    c1#is_solved || c2#is_solved || 
    try
      if not c1#check then begin (* if c1 is false, c2 must be true *)
	Fcl_cstr.post c2
      end;
      true
    with
      Fcl_cstr.DontKnow -> 
	try
	  if not c2#check then (* if c2 is false, c1 must be true *)
	    Fcl_cstr.post c1;
	  true
	with
	  Fcl_cstr.DontKnow -> false

  and fprint s = Printf.fprintf s "("; c1#fprint s; Printf.fprintf s ") ||~~ ("; c2#fprint s; Printf.fprintf s ")"

  and delay c =
    c1#delay c;
    c2#delay c;
    c1#not#delay c;
    c2#not#delay c

  and check () =
     c1#is_solved || c2#is_solved ||
    try 
      (try c1#check with Fcl_cstr.DontKnow -> raise MyDontKnow) || c2#check
    with
      MyDontKnow ->
	c2#check || raise Fcl_cstr.DontKnow

  and not () = (&&~~) c1#not c2#not

  in

  Fcl_cstr.create ~name:"||~~" ~fprint:fprint ~not:not ~check:check update delay

and (&&~~) c1 c2 =
  let update () =
    Fcl_cstr.post c1;
    Fcl_cstr.post c2;
    true

  and fprint s = Printf.fprintf s "("; c1#fprint s; Printf.fprintf s ") &&~~ ("; c2#fprint s; Printf.fprintf s ")"

  and delay c =
    c1#delay c;
    c2#delay c;
    c1#not#delay c;
    c2#not#delay c

  and check () =
    (c1#is_solved ||
      try c1#check with
	Fcl_cstr.DontKnow ->
	  if c2#check then raise Fcl_cstr.DontKnow else false )
      &&
    (c2#is_solved || c2#check)

  and not () = (||~~) c1#not c2#not

  in

  Fcl_cstr.create ~name:"&&~~" ~fprint:fprint ~not:not ~check:check update delay;;

let (=>~~) c1 c2 = c1#not ||~~ c2

let rec eq_or_xor c1 c2 equiv = (* if [equiv] then (<=>~~) else (xor~~) *)
  let update () =
    try
      Fcl_cstr.post (if c1#check = equiv then c2 else c2#not);
      true
    with
      Fcl_cstr.DontKnow -> (* c1 unknown *)
	try
	  Fcl_cstr.post (if c2#check = equiv then c1 else c1#not);
	  true
	with
	  Fcl_cstr.DontKnow -> (* c1 && c2 unknown *)
	    false

  and delay c =
    c1#delay c;
    c2#delay c;
    c1#not#delay c;
    c2#not#delay c

  and check () =
    (c1#check = c2#check) = equiv

  and not () =
    eq_or_xor c1 c2 (not equiv)

  in

  Fcl_cstr.create ~name:(if equiv then "<=>~~" else "xor~~") ~not:not ~check:check update delay;;


let (<=>~~) c1 c2 = eq_or_xor c1 c2 true

let xor c1 c2 = eq_or_xor c1 c2 false

let not c = c#not
